#!/usr/bin/python

## expendable

## Copyright (C) 2008, 2009 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import sys
if __name__ == '__main__':
    sys.path.append (os.environ.get ('top_srcdir', '.'))

from gettext import gettext as _
import gtk.glade
import locale
import account

name=_("Fixed Term Deposit")
description=_("Fixed term deposit with taxable interest paid on maturity.")

class Model(account.Model):
    def __init__ (self, rate, taxrate, amount, term):
        account.Model.__init__ (self)
        self.rate = rate
        self.taxrate = taxrate
        self.amount = amount
        self.balance = 0.0
        self.term = term
        self.gross = 0
        self.net = 0

    def cash_available (self):
        if self.per > self.term:
            return self.balance

        return 0

    def append_data (self, month, gross=0.0, net=0.0, tax=0.0):
        self.model_data.append( { 'month': month,
                                  'balance': self.balance,
                                  'totalgross': self.gross,
                                  'totalnet': self.net,
                                  'totaltax': self.gross - self.net,
                                  'gross': gross,
                                  'net': net,
                                  'tax': tax } )
    def iter (self):
        if self.per == 0:
            if self.parent != None:
                contrib = self.parent.adjust (-self.amount, self)
                if contrib < self.amount:
                    raise account.ModelError (self, "Not enough for FTD")

            self.balance = self.amount

        if self.per == self.term:
            gross = int (100 * self.balance *
                         self.rate * self.term / 12.0) * 0.01
            tax = int (100 * gross * self.taxrate) * 0.01
            net = gross - tax
            self.balance += net
            self.gross += gross
            self.net += net
            self.append_data (month=self.per, gross=gross, net=net, tax=tax)
        else:
            self.append_data (month=self.per)

        if self.parent != None and self.per == self.term:
            self.parent.adjust (self.balance, self)
            self.balance = 0.0
            self.append_data (month=self.per + 1)
            return True

        return False

    def get_datasets (self):
        model_data = self.model_data
        sdata = map (lambda x: (x['month'], x['balance']), model_data)
        return [{'name': self.account.get_name (),
                 'data': sdata}]

class Data(account.Data):
    required_tags = ['rate', 'taxrate', 'amount', 'term']
    glade = "ftd.glade"

    def __init__ (self):
        account.Data.__init__ (self)
        for dir in sys.path:
            f = os.path.join (dir, self.glade)
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        data = {}
        for name in ['rate', 'taxrate']:
            widget = gld.get_widget (name)
            data[name] = widget.get_value () / 100.0

        for name in ['amount', 'term']:
            widget = gld.get_widget (name)
            data[name] = widget.get_value ()

        data['term'] = int (data['term'])
        self.data = data

    def is_leaf_node (self):
        return True

    def read (self, xml):
        data = {}
        for child in xml.getchildren ():
            # Floats
            if child.tag in ['rate',
                             'taxrate',
                             'amount']:
                data[child.tag] = float (child.text)
            elif child.tag in ['term']:
                # Ints
                data[child.tag] = int (child.text)
            elif child.tag == 'balance':
                # Convert old name for this tag.
                data['amount'] = float (child.text)
            else:
                print "Not understood: %s" % child.tag

        for required in self.required_tags:
            if not data.has_key (required):
                raise RuntimeError, "Required tags: %s" % self.required_tags

        self.data = data

    def write (self):
        xml = "<ftd>\n"
        for var in self.required_tags:
            value = self.data[var]
            if type (value) == float:
                value = "%.4f" % value
            xml += "  <%s>%s</%s>\n" % (var, value, var)
        xml += "</ftd>\n"
        return xml

    def ui (self, variables_vbox, results_vbox):
        for dir in sys.path:
            f = os.path.join (dir, self.glade)
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        w = gld.get_widget ('variables_table')
        w.reparent (variables_vbox)
        for var in ['rate', 'taxrate']:
            widget = gld.get_widget (var)
            widget.set_value (self.data[var] * 100.0)
            widget.connect ("value-changed", self.rate_changed, var)
        for var in ['amount', 'term']:
            widget = gld.get_widget (var)
            widget.set_value (self.data[var])
            widget.connect ("value-changed", self.variable_changed, var)

        w = gld.get_widget ('results_hbox')
        w.reparent (results_vbox)
        self.result_widget = {}
        for name in ['fbalance', 'totalnet']:
            self.result_widget[name] = gld.get_widget (name)

    def rate_changed (self, widget, name):
        self.data[name] = widget.get_value () / 100.0
        try:
            self.redisplay ()
        except:
            widget.set_value (self.last_known_good[name] * 100.0)

    def variable_changed (self, widget, name):
        value = widget.get_value ()
        if name == 'term':
            value = int (value)
        self.data[name] = value
        try:
            self.redisplay ()
        except:
            widget.set_value (self.last_known_good[name])

    def update_results (self, modeldata):
        fbalance = self.result_widget['fbalance']
        totalnet = self.result_widget['totalnet']
        n = len (modeldata)
        if n > 0:
            last = modeldata[n - 1]
            fbalance.set_text (locale.currency (self.data['amount'] +
                                                last['totalnet'], True, True))
            totalnet.set_text (locale.currency (last['totalnet'], True, True))
        else:
            fbalance.set_text ('')
            totalnet.set_text ('')

    def get_model (self):
        return Model (**self.data)

if __name__ == '__main__':
    from pprint import *
    from xml.etree.ElementTree import XML
    d = Data ()
    w = d.write ()
    xml = XML (w)
    d.read (xml)

    import savings
    s = savings.Model (rate=0,
                       balance=5000,
                       taxrate=0)
    ftd = Model (rate=0.10,
                 amount=5000,
                 taxrate=0.20,
                 term=12)
    s.add_model (ftd)
    s.run (nper=13)
    data = ftd.get_model_data ()
    #pprint (data)

    # We expect the initial month, 12 months of interest, then a final
    # 'zero' month.
    if len (data) != 14:
        raise RuntimeError, "FTD under/overrun"

    data = s.get_model_data ()
    first = data[0]['balance']
    penult = data[12]['balance']
    final = data[13]['balance']
    #pprint (data)
    #print final

    if abs (first - 5000.00) > 0.01:
        raise RuntimeError, "Incorrect initial balance"

    if penult >= 0.01:
        raise RuntimeError, "FTD failed to take savings for initial balance"

    if abs (final - 5400.00) > 0.01:
        raise RuntimeError, "Incorrect FTD return"
