#!/usr/bin/python

## expendable

## Copyright (C) 2008 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import errno
import gconf
import getopt
from gettext import gettext as _
import gettext
import glob
import gobject
import gtk.glade
import locale
import os.path
import sys

pkgdatadir = os.path.dirname (sys.argv[0])
sys.path.append (pkgdatadir)
accountsdir = os.path.join (pkgdatadir, "accounts")
sys.path.append (accountsdir)
locale.setlocale (locale.LC_ALL, "")
gettext.textdomain ("expendable")
from graph import Graph
import xmlfile

example_file=os.path.join (pkgdatadir, "example.xml")

class Main:
    KEY_LAST_FILENAME="/apps/expendable/last-filename"
    KEY_WINDOW_WIDTH="/apps/expendable/window-width"
    KEY_WINDOW_HEIGHT="/apps/expendable/window-height"

    def __init__ (self, initial_file=None):
        gtk.window_set_default_icon_name ('expendable')
        for dir in sys.path:
            f = os.path.join (dir, "window.glade")
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        self.account_name_has_changed = False
        self.window = gld.get_widget ('window')
        self.window.connect ('delete_event', self.destroy)

        self.graph = Graph ()
        sw = gld.get_widget ('graphviewport')
        sw.add (self.graph)

        self.gconf_client = gconf.Client ()
        self.run_model = gld.get_widget ('run_model')
        self.run_model.connect ("toggled", self.run_model_toggled)
        self.limit_years = gld.get_widget ('limit_years')
        self.limit_months = gld.get_widget ('limit_months')
        self.limit_years.connect ("value-changed", self.limit_changed)
        self.limit_months.connect ("value-changed", self.limit_changed)

        treeview = gld.get_widget ('treeview')
        self.accounts_tree_model = gtk.TreeStore (gobject.TYPE_STRING,
                                                  gobject.TYPE_STRING,
                                                  gobject.TYPE_INT)
        treeview.set_model (self.accounts_tree_model)
        col = gtk.TreeViewColumn (_("Name"),
                                  gtk.CellRendererText (),
                                  text=0)
        col.set_sizing (gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        treeview.append_column (col)
        col = gtk.TreeViewColumn (_("Type"),
                                  gtk.CellRendererText (),
                                  text=1)
        treeview.append_column (col)
        self.treeview = treeview
        self.treeview.connect ("cursor-changed", self.treeview_cursor_changed)

        self.notebook = gld.get_widget ('notebook')
        self.results_notebook = gld.get_widget ('results_notebook')

        new = gld.get_widget ("file_new")
        new.connect ("activate", self.file_new)
        open = gld.get_widget ("file_open")
        open.connect ("activate", self.file_open)
        save = gld.get_widget ("file_save")
        save.connect ("activate", self.file_save)
        save.set_sensitive (False)
        save_as = gld.get_widget ("file_save_as")
        save_as.connect ("activate", self.file_save_as)
        save_as.set_sensitive (False)
        self.save = save
        self.save_as = save_as
        quit = gld.get_widget ("file_quit")
        quit.connect ("activate", self.destroy)

        delete_account = gld.get_widget ("edit_delete_account")
        delete_account.connect ("activate", self.edit_delete_account)
        self.delete_account = delete_account

        add_account = gld.get_widget ("edit_add_account")
        add_account.connect ("activate", self.edit_add_account)

        self.intin_label = gld.get_widget ("intin")
        self.intout_label = gld.get_widget ("intout")
        self.intnet_label = gld.get_widget ("intnet")

        for dir in sys.path:
            f = os.path.join (dir, "add-account.glade")
            if os.access (f, os.R_OK):
                addgld = gtk.glade.XML (f)
                break

        add_account_dialog = addgld.get_widget ('add_account_dialog')
        add_account_dialog.connect ("delete-event", self.glade_delete_event)
        add_account_dialog.connect ("close", self.edit_add_account_close)
        add_account_dialog.connect ("response",
                                    self.edit_add_account_response)
        self.add_account_dialog = add_account_dialog
        self.add_account_description = addgld.get_widget ('account_description')
        self.add_account_type = addgld.get_widget ('account_type')
        col = gtk.TreeViewColumn (_("Account Type"),
                                  gtk.CellRendererText (),
                                  text=0)
        col.set_sort_column_id (0)
        self.add_account_type.append_column (col)
        self.add_account_type.connect ("cursor-changed",
                                       self.edit_add_account_type_changed)
        self.add_account_type.connect ("row-activated",
                                       self.add_account_type_row_activated)

        open_filename = example_file

        last_filename = self.gconf_client.get_string (self.KEY_LAST_FILENAME)
        if last_filename != None and last_filename != '':
            open_filename = last_filename

        if initial_file != None:
            open_filename = initial_file

        (width, height) = (self.gconf_client.get_int (self.KEY_WINDOW_WIDTH),
                           self.gconf_client.get_int (self.KEY_WINDOW_HEIGHT))
        self.window.set_default_size (width, height)
        self.window.show_all ()
        self.window.connect ("check-resize", self.on_check_resize)

        remember_file = True
        while True:
            try:
                self.load_file (open_filename, remember_file=remember_file)
            except IOError, e:
                if e.errno == errno.EAGAIN:
                    remember_file = False

                if open_filename != example_file:
                    open_filename = example_file
                    continue

            break

    def on_check_resize (self, widget):
        (width, height) = widget.get_size ()
        self.gconf_client.set_int (self.KEY_WINDOW_WIDTH, width)
        self.gconf_client.set_int (self.KEY_WINDOW_HEIGHT, height)

    def glade_delete_event (self, widget, event):
        widget.hide ()
        return True

    def set_current_filename (self, filename):
        if self.check_if_changes_made ():
            changes = '*'
        else:
            changes = ''

        if filename == None:
            filename = _("Untitled")

        title = _("%s%s - expendable") % (changes, filename)
        if self.xmlfile.is_readonly ():
            title += " " + _("(read-only)")

        self.window.set_title (title)

    def run_model_toggled (self, widget):
        if widget.get_active ():
            self.graph.show ()
            self.results_notebook.show_all ()
            try:
                self.redisplay ()
            except:
                pass
        else:
            self.graph.hide ()
            self.results_notebook.hide ()

    def limit_changed (self, widget):
        years = self.limit_years.get_value ()
        months = self.limit_months.get_value ()
        if years == 0 and months == 0:
            if widget == self.limit_years:
                years = 1
                self.limit_years.set_value (years)
            else:
                months = 1
                self.limit_months.set_value (months)
        elif months < 0:
            months = 12 + months
            years -= 1
            self.limit_months.set_value (months)
            self.limit_years.set_value (years)
        elif months > 11:
            months = months - 12
            years += 1
            self.limit_months.set_value (months)
            self.limit_years.set_value (years)

        self.xmlfile.set_limit_nper (int (12 * years + months))
        self.redisplay ()

    def file_new (self, widget):
        if self.check_if_changes_made ():
            response = self.maybe_save_changes ()
            if response == gtk.RESPONSE_CANCEL:
                return

        toplevel_types = []
        types = map (lambda x: os.path.splitext (os.path.basename (x))[0],
                     glob.glob (os.path.join (accountsdir, "*.py")))
        for type in types:
            data = xmlfile.create_account (type, '')
            if data.is_leaf_node ():
                continue
            toplevel_types.append ((data.get_type (),
                                    data.get_description (),
                                    type))

        types = toplevel_types
        model = gtk.ListStore (gobject.TYPE_STRING,
                               gobject.TYPE_STRING,
                               gobject.TYPE_STRING)
        for n, d, t in types:
            model.append ([n, d, t])
        treeview = self.add_account_type
        treeview.set_model (model)
        treeview.set_cursor ((0,))
        self.add_account_dialog.set_data ('new', True)
        self.add_account_dialog.set_transient_for (self.window)
        self.add_account_dialog.show ()

    def file_open (self, widget):
        if self.check_if_changes_made ():
            response = self.maybe_save_changes ()
            if response == gtk.RESPONSE_CANCEL:
                return

        fs = gtk.FileChooserDialog (_("Choose File"), self.window,
                                    gtk.FILE_CHOOSER_ACTION_OPEN,
                                    (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                     gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        fs.connect ("response", self.file_open_response)
        fs.connect ("close", self.file_open_close)
        fs.show ()

    def file_open_close (self, widget):
        widget.destroy ()

    def file_open_response (self, widget, response):
        if response != gtk.RESPONSE_OK:
            widget.destroy ()
            return

        f = widget.get_filename ()
        widget.destroy ()
        try:
            self.load_file (f)
        except IOError:
            # Was handled by a dialog in load_file.
            pass

    def file_save (self, widget, new_filename=None):
        if new_filename == None and self.xmlfile.get_filename () == None:
            return self.file_save_as (widget)

        try:
            if new_filename == None:
                self.xmlfile.save ()
            else:
                self.xmlfile.save (new_filename=new_filename)

            self.original_xml = self.xmlfile.get_xml ()
            self.set_current_filename (self.xmlfile.get_filename ())
        except Exception, e:
            d = gtk.MessageDialog (self.window,
                                   gtk.DIALOG_DESTROY_WITH_PARENT,
                                   gtk.MESSAGE_ERROR,
                                   gtk.BUTTONS_CLOSE)
            d.set_markup ('<span weight="bold" size="larger">' +
                          _("Save Failed") + '</span>\n\n' +
                          _("Failed to save file.") + " " +
                          _("The error was:") + "\n\n" + str (e))
            d.run ()
            d.destroy ()
            return

    def file_save_as (self, widget=None):
        fs = gtk.FileChooserDialog (_("Choose Filename"), self.window,
                                    gtk.FILE_CHOOSER_ACTION_SAVE,
                                    (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                     gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        fs.set_do_overwrite_confirmation (True)
        fs.connect ("response", self.file_save_as_response)
        fs.connect ("close", self.file_save_as_close)
        fs.show ()
        self.file_chooser_dialog = fs

    def file_save_as_close (self, widget):
        widget.destroy ()

    def file_save_as_response (self, widget, response):
        if response != gtk.RESPONSE_OK:
            widget.destroy ()
            return

        f = widget.get_filename ()
        self.file_save (widget, new_filename=f)
        self.original_xml = self.xmlfile.get_xml ()
        self.set_current_filename (f)
        widget.destroy ()

    def edit_delete_account (self, widget):
        treeview = self.treeview
        path, col = treeview.get_cursor ()
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            parent = accounts[path[step]]
            accounts = parent.accounts
        account = accounts[path[len (path) - 1]]
        index = parent.accounts.index (account)
        del parent.accounts[index]
        gobject.timeout_add (0, self.new_account_tree)

    def edit_add_account (self, widget):
        possible_types = []
        types = map (lambda x: os.path.splitext (os.path.basename (x))[0],
                     glob.glob (os.path.join (accountsdir, "*.py")))
        for type in types:
            data = xmlfile.create_account (type, '')
            possible_types.append ((data.get_type (),
                                    data.get_description (),
                                    type))
        types = possible_types
        model = gtk.ListStore (gobject.TYPE_STRING,
                               gobject.TYPE_STRING,
                               gobject.TYPE_STRING)
        model.set_sort_column_id (0, gtk.SORT_ASCENDING)
        for n, d, t in types:
            model.append ([n, d, t])
        treeview = self.add_account_type
        treeview.set_model (model)
        treeview.set_cursor ((0,))
        self.add_account_dialog.set_data ('new', False)
        self.add_account_dialog.set_transient_for (self.window)
        self.add_account_dialog.show ()
        treeview.grab_focus ()

    def edit_add_account_type_changed (self, widget):
        path, col = widget.get_cursor ()
        model = widget.get_model ()
        iter = model.get_iter (path)
        description = model.get_value (iter, 1)
        self.add_account_description.set_text (description)

    def add_account_type_row_activated (self, widget, path, column):
        self.add_account_dialog.response (gtk.RESPONSE_OK)

    def edit_add_account_close (self, widget):
        widget.hide ()

    def edit_add_account_response (self, widget, response):
        if response != gtk.RESPONSE_OK:
            widget.hide ()
            return

        treeview = self.add_account_type
        model = treeview.get_model ()
        path, col = treeview.get_cursor ()
        iter = model.get_iter (path)
        type = model.get_value (iter, 2)
        account = xmlfile.create_account (type, _("New Account"))
        account._set_last_known_good ()
        if self.add_account_dialog.get_data ('new'):
            self.last_error_per = None
            self.xmlfile = xmlfile.File (toplevel=account)
            self.accounts = self.xmlfile.get_accounts ()
            self.original_xml = ''
        else:
            # Find out which account was already selected.
            path, col = self.treeview.get_cursor ()
            acs = self.accounts
            for step in xrange (len (path) - 1):
                parent = acs[path[step]]
                acs = parent.accounts
            ac = acs[path[len (path) - 1]]
            if path == (0,):
                # If the top-level account was selected, link the new
                # account to that.
                ac.accounts.insert (0, account)
            elif ac.is_leaf_node ():
                # If the selected account is a leaf node, link the new
                # account to its parent.  The new account is now a
                # sibling of the previously selected account.
                parent.accounts.insert (parent.accounts.index (ac),
                                        account)
            else:
                # Otherwise, link the new account to the selected account.
                ac.accounts.insert (0, account)

        widget.hide ()
        self.new_account_tree (account)
        if self.add_account_dialog.get_data ('new'):
            limit_nper = self.xmlfile.get_limit_nper ()
            self.limit_years.set_value (limit_nper / 12)
            self.limit_months.set_value (limit_nper % 12)

    def load_file (self, filename, remember_file=True):
        excp = None
        try:
            self.xmlfile = xmlfile.File (filename)
        except IOError, e:
            if e.errno == errno.EAGAIN:
                excp = _("The file is locked by another application.")
            else:
                excp = _("The error was:") + "\n\n" + str (e)
        except Exception, e:
            excp = _("The error was:") + "\n\n" + str (e)

        if excp != None:
            d = gtk.MessageDialog (self.window,
                                   gtk.DIALOG_DESTROY_WITH_PARENT,
                                   gtk.MESSAGE_ERROR,
                                   gtk.BUTTONS_CLOSE)
            d.set_markup ('<span weight="bold" size="larger">' +
                          _("Open Failed") + '</span>\n\n' +
                          _("Failed to open file '%s'.") % filename +
                          " " + excp)
            d.run ()
            d.destroy ()
            raise

        self.original_xml = self.xmlfile.get_xml ()
        self.last_error_per = None
        self.save.set_sensitive (not self.xmlfile.is_readonly ())
        self.save_as.set_sensitive (True)
        self.accounts = self.xmlfile.get_accounts ()
        self.update_accounts_tree (self.accounts)
        limit_nper = self.xmlfile.get_limit_nper ()
        self.limit_years.set_value (limit_nper / 12)
        self.limit_months.set_value (limit_nper % 12)
        self.redisplay ()

        if remember_file and filename[0] == '/':
            gc = gconf.Client ()
            gc.set_string (self.KEY_LAST_FILENAME, filename)

        def set_last_known_good (accounts):
            for account in accounts:
                account._set_last_known_good ()
                set_last_known_good (account.accounts)

        set_last_known_good (self.accounts)

    def check_if_changes_made (self):
        try:
            xml = self.xmlfile.get_xml ()
            return xml != self.original_xml
        except AttributeError:
            # Not loaded yet
            return False

    def update_accounts_tree (self, accounts):
        self.accounts_tree_model.clear ()
        for n in xrange (self.notebook.get_n_pages ()):
            self.notebook.remove_page (0)

        self.results_notebook.hide ()
        for n in xrange (self.results_notebook.get_n_pages () - 1):
            self.results_notebook.remove_page (1)

        def find_branch_nodes (accounts):
            nodes = []
            for account in accounts:
                if not account.is_leaf_node ():
                    nodes.append (account)
                    nodes.extend (find_branch_nodes (account.accounts))
            return nodes

        branch_nodes = find_branch_nodes (accounts)
        name_widgets = {}

        def add_accounts (accounts, parent, path=None, parent_account=None):
            path_n = 0
            num_accounts = len (accounts)
            for index in xrange (num_accounts):
                account = accounts[index]
                if parent == None:
                    this_path = (path_n,)
                else:
                    this_path = path + (path_n,)

                vbox = gtk.VBox ()
                vbox.set_spacing (6)

                table = gtk.Table (1, 2)
                vbox.pack_start (table, False, False, 0)
                table.set_col_spacings (6)
                table.set_row_spacings (6)
                row = 0

                hbox = gtk.HBox ()
                hbox.set_spacing (12)
                table.attach (hbox, 0, 2, row, row + 1)

                bbox = gtk.HButtonBox ()
                bbox.set_layout (gtk.BUTTONBOX_START)
                hbox.pack_start (bbox)
                up = gtk.Button (stock=gtk.STOCK_GO_UP)
                bbox.pack_start (up)
                down = gtk.Button (stock=gtk.STOCK_GO_DOWN)
                bbox.pack_end (down)
                if index == 0:
                    up.set_sensitive (False)
                if index == num_accounts - 1:
                    down.set_sensitive (False)

                up.connect ("clicked", self.account_up_clicked, this_path)
                down.connect ("clicked", self.account_down_clicked, this_path)

                cb = gtk.CheckButton (_("Plot"))
                hbox.pack_start (cb, False, False, 0)
                cb.set_active (account.get_plot ())
                cb.connect ("toggled", self.account_plot_toggled, this_path)
                account.plot_cb = cb

                row += 1

                label = gtk.Label (_("Account name:"))
                label.set_alignment (0.0, 0.5)
                table.attach (label, 0, 1, row, row + 1)

                entry = gtk.Entry ()
                entry.set_text (account.get_name ())
                entry.connect ("changed", self.account_name_changed, this_path)
                entry.connect ("activate",
                               self.account_name_activated, this_path)
                account.name_entry = entry
                table.attach (entry, 1, 2, row, row + 1)
                name_widgets[this_path] = entry
                row += 1

                if parent != None:
                    available_branch_nodes = set (branch_nodes)
                    try:
                        for sub in find_branch_nodes (account.accounts):
                            available_branch_nodes.remove (sub)
                        available_branch_nodes.remove (account)
                    except KeyError:
                        pass

                    if len (available_branch_nodes) > 1:
                        label = gtk.Label (_("Linked account:"))
                        label.set_alignment (0.0, 0.5)
                        table.attach (label, 0, 1, row, row + 1)
                        model = gtk.ListStore (gobject.TYPE_STRING,
                                               gobject.TYPE_PYOBJECT)
                        current = 0
                        i = 0
                        for node in available_branch_nodes:
                            model.append ([node.get_name (), node])
                            if parent_account == node:
                                current = i
                            i += 1
                        combobox = gtk.ComboBox (model=model)
                        cell = gtk.CellRendererText ()
                        combobox.pack_start (cell, True)
                        combobox.add_attribute (cell, 'text', 0)
                        combobox.set_active (current)
                        combobox.connect ("changed",
                                          self.linked_account_changed,
                                          this_path)
                        table.attach (combobox, 1, 2, row, row + 1)
                        row += 1

                    label = gtk.Label (_("Start:"))
                    label.set_alignment (0.0, 0.5)
                    table.attach (label, 0, 1, row, row + 1)
                    hbox = gtk.HBox ()
                    hbox.set_spacing (6)
                    table.attach (hbox, 1, 2, row, row + 1)
                    adj = gtk.Adjustment (value=0, lower=0, upper=240,
                                          step_incr=1)
                    spinbutton = gtk.SpinButton (adj)
                    delay = account.get_delay ()
                    spinbutton.set_value (abs (delay))
                    spinbutton.connect ("value-changed",
                                        self.early_late_changed,
                                        this_path)
                    account.delay_spinbutton = spinbutton
                    hbox.pack_start (spinbutton, False, False, 0)
                    label = gtk.Label (_("mths"))
                    label.set_alignment (0.0, 0.5)
                    hbox.pack_start (label, False, False, 0)
                    model = gtk.ListStore (gobject.TYPE_STRING,
                                           gobject.TYPE_INT)
                    model.append (row=[_("early"), -1])
                    model.append (row=[_("late"), 1])
                    combobox = gtk.ComboBox (model=model)
                    hbox.pack_start (combobox, False, False, 0)
                    cell = gtk.CellRendererText ()
                    combobox.pack_start (cell, True)
                    combobox.add_attribute (cell, 'text', 0)
                    if delay >= 0:
                        combobox.set_active (1)
                    else:
                        combobox.set_active (0)
                    account.delay_combobox = combobox
                    combobox.connect ("changed",
                                      self.early_late_changed,
                                      this_path)
                    row += 1

                results_vbox = gtk.VBox ()
                results_vbox.set_border_width (12)
                account.ui (vbox, results_vbox)
                self.notebook.append_page (vbox)
                name = account.get_name ()
                if len (results_vbox.get_children ()) > 0:
                    label = gtk.Label (name)
                    self.results_notebook.append_page (results_vbox,
                                                       tab_label=label)
                    self.results_notebook.show_all ()
                n = self.notebook.get_n_pages () - 1
                type = account.get_type ()
                iter = self.accounts_tree_model.append (parent, [name, type, n])
                account.set_redisplay_fn (self.redisplay)
                add_accounts (account.accounts, iter, path=this_path,
                              parent_account=account)
                path_n += 1

        add_accounts (accounts, None)
        self.name_widgets = name_widgets
        self.treeview.expand_all ()
        self.treeview.set_cursor ((0,))
        self.notebook.show_all ()

    def treeview_cursor_changed (self, treeview):
        path, col = treeview.get_cursor ()
        if path == None:
            return
        model = treeview.get_model ()
        iter = model.get_iter (path)
        page = model.get_value (iter, 2)
        self.notebook.set_current_page (page)
        self.delete_account.set_sensitive (path != (0,))

    def account_up_clicked (self, widget, path):
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            old_parent = accounts[path[step]]
            accounts = old_parent.accounts
        index = path[len (path) - 1]
        account = accounts[index]
        del old_parent.accounts[index]
        old_parent.accounts.insert (index - 1, account)
        gobject.timeout_add (0, self.new_account_tree, account)

    def account_down_clicked (self, widget, path):
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            old_parent = accounts[path[step]]
            accounts = old_parent.accounts
        index = path[len (path) - 1]
        account = accounts[index]
        del old_parent.accounts[index]
        old_parent.accounts.insert (index + 1, account)
        gobject.timeout_add (0, self.new_account_tree, account)

    def account_plot_toggled (self, widget, path):
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            accounts = accounts[path[step]].accounts
        account = accounts[path[len (path) - 1]]
        account.set_plot (widget.get_active ())
        self.redisplay ()

    def account_name_changed (self, widget, path):
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            accounts = accounts[path[step]].accounts
        account = accounts[path[len (path) - 1]]
        entry = account.name_entry
        name = entry.get_text ()
        account.set_name (name)
        model = self.accounts_tree_model
        iter = model.get_iter (path)
        model.set_value (iter, 0, name)
        self.account_name_has_changed = True

    def account_name_activated (self, widget, path):
        treeview = self.treeview
        path, col = treeview.get_cursor ()
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            parent = accounts[path[step]]
            accounts = parent.accounts
        account = accounts[path[len (path) - 1]]
        if self.account_name_has_changed:
            self.account_name_has_changed = False
            gobject.timeout_add (0, self.new_account_tree, account)

    def linked_account_changed (self, widget, path):
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            old_parent = accounts[path[step]]
            accounts = old_parent.accounts
        account = accounts[path[len (path) - 1]]
        iter = widget.get_active_iter ()
        model = widget.get_model ()
        new_parent = model.get_value (iter, 1)

        index = old_parent.accounts.index (account)
        del old_parent.accounts[index]
        new_parent.accounts.insert (0, account)
        gobject.timeout_add (0, self.new_account_tree, account)

    def new_account_tree (self, account=None):
        self.update_accounts_tree (self.accounts)
        def get_path (accounts, path=[]):
            for step in xrange (len (accounts)):
                this_path = path + [step]
                if accounts[step] == account:
                    return tuple (this_path)
                p = get_path (accounts[step].accounts, this_path)
                if p != None:
                    return p
            return None
        if account != None:
            path = get_path (self.accounts)
        else:
            path = (0,)
        self.treeview.set_cursor (path)
        self.treeview.scroll_to_cell (path,
                                      use_align=True,
                                      row_align=0.5)
        entry = self.name_widgets[path]
        entry.grab_focus ()
        try:
            self.redisplay ()
        except:
            pass

    def early_late_changed (self, widget, path):
        accounts = self.accounts
        for step in xrange (len (path) - 1):
            accounts = accounts[path[step]].accounts
        account = accounts[path[len (path) - 1]]
        cb = account.delay_combobox
        sb = account.delay_spinbutton
        iter = cb.get_active_iter ()
        model = cb.get_model ()
        m = model.get_value (iter, 1)
        by = int (sb.get_value ())
        delay = m * by
        try:
            account.set_delay (delay)
            self.redisplay ()
            account.last_known_good['_delay'] = delay
        except:
            delay = account.last_known_good['_delay']
            try:
                if delay >= 0:
                    cb.set_active (1)
                else:
                    cb.set_active (0)
            except:
                pass

            try:
                sb.set_value (abs (delay))
            except:
                pass

    def build_models (self, accounts):
        models = []
        for account in accounts:
            model = account.get_model ()
            model.account = account
            models.append (model)
            submodels = self.build_models (account.accounts)
            for submodel in submodels:
                model.add_model (submodel)

            model.set_delay (account.get_delay ())

        return models

    def redisplay (self):
        self.set_current_filename (self.xmlfile.get_filename ())

        if not self.run_model.get_active ():
            return

        limit = self.limit_years.get_value () * 12
        limit += self.limit_months.get_value ()
        exception = None
        models = self.build_models (self.accounts)
        for model in models:
            try:
                model.run (nper=limit + 1)
            except Exception, e:
                exception = e
                self.last_error_per = models[0].per
                break

        colors = ['red', 'blue', 'green', 'cyan', 'orange',
                  'magenta', 'black', 'purple', 'yellow']
        self.intin = 0.0
        self.intout = 0.0
        def get_datasets (models):
            datasets = []
            for model in models:
                data = model.get_model_data ()
                model.account.update_results (data)
                nper = len (data)
                if nper > 0:
                    last = data[len (data) - 1]
                    self.intout += last.get ('total interest', 0.0)
                    self.intin += last.get ('totalnet', 0.0)

                if model.account.get_plot ():
                    sets = model.format_datasets ()
                    model.account.plot_cb.set_sensitive (len (sets) != 0)

                    for set in sets:
                        try:
                            chosen = set['color']
                            try:
                                del colors[colors.index (chosen)]
                            except KeyError:
                                pass
                        except KeyError:
                            # Assign a color automatically.
                            try:
                                set['color'] = colors.pop (0)
                            except IndexError:
                                # Out of colors.
                                set['color'] = 'black'

                    datasets.extend (sets)

                datasets.extend (get_datasets (model.get_models ()))
            return datasets

        if self.last_error_per != None:
            self.graph.set_datasets (get_datasets (models),
                                     self.last_error_per)
        else:
            self.graph.set_datasets (get_datasets (models))

        self.intin_label.set_text (locale.currency (self.intin, True, True))
        self.intout_label.set_text (locale.currency (self.intout, True, True))
        net = self.intin - self.intout
        nettext = locale.currency (net, True, True)
        if net < 0.0:
            nettext = "<span color='red'>%s</span>" % nettext
        self.intnet_label.set_markup (nettext)

        if exception != None:
            print "At month %s: %s" % (models[0].per, exception)
            import pprint
            def dump_model_data (models):
                for model in models:
                    print
                    print model
                    pprint.pprint (model.get_model_data ())
                    dump_model_data (model.get_models ())
            if False:
                import traceback
                (type, value, tb) = sys.exc_info ()
                tblast = traceback.extract_tb (tb, limit=None)
                if len (tblast):
                    tblast = tblast[:len (tblast) - 1]
                extxt = traceback.format_exception_only (type, value)
                for line in traceback.format_tb(tb):
                    print line.strip ()
                print extxt[0].strip ()
            if False:
                dump_model_data (models)
            raise exception
        else:
            self.last_error_per = None

        return

    def maybe_save_changes (self):
        m = gtk.MessageDialog (self.window,
                               gtk.DIALOG_MODAL |
                               gtk.DIALOG_DESTROY_WITH_PARENT,
                               gtk.MESSAGE_QUESTION,
                               gtk.BUTTONS_NONE)
        m.add_buttons (_("Discard changes"), gtk.RESPONSE_NO,
                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                       gtk.STOCK_SAVE, gtk.RESPONSE_YES)
        filename = self.xmlfile.get_filename ()
        requires_filename = (filename == None)
        if requires_filename:
            filename = _("Untitled")
        m.set_markup ('<span size="larger" weight="bold">' +
                      _("Save changes to %s?") % filename +
                      '</span>')
        m.set_default_response (gtk.RESPONSE_YES)
        response = m.run ()
        m.destroy ()
        if response == gtk.RESPONSE_YES:
            if requires_filename:
                self.file_save_as ()
                response = self.file_chooser_dialog.run ()
                self.file_save_as_response (self.file_chooser_dialog, response)
            else:
                self.xmlfile.save ()

        return response

    def destroy (self, widget, *args):
        if self.check_if_changes_made ():
            response = self.maybe_save_changes ()
            if response == gtk.RESPONSE_CANCEL:
                return True
        gtk.main_quit ()

opts, args = getopt.getopt (sys.argv[1:], "")
params = dict()
if len (args) >= 1:
    params['initial_file'] = args[0]

m = Main (**params)
gtk.main ()
