#!/usr/bin/python

## expendable

## Copyright (C) 2008 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import errno
import fcntl
from xml.etree.ElementTree import XML, tostring

def create_account (type, name):
    exec "import %s" % type
    exec "data = %s.Data ()" % type
    exec "tname = %s.name" % type
    exec "description = %s.description" % type
    data.set_type (tname)
    data.set_name (name)
    data.set_description (description)
    return data

class LockedFile(file):
    def __init__ (self, filename, mode='r'):
        if mode == 'r':
            mode += '+'

        file.__init__ (self, filename, mode)
        fcntl.lockf (self, fcntl.LOCK_EX | fcntl.LOCK_NB)

class File:
    def __init__ (self, filename=None, toplevel=None):
        if filename != None:
            try:
                self.file = LockedFile (filename)
            except IOError, e:
                if e.errno == errno.EACCES:
                    self.file = file (filename)
                else:
                    raise

            xml = XML (self.file.read ())
            for node in xml.getchildren ():
                if node.tag == 'accounts':
                    self.limit_nper = int (node.attrib.get ("limit-nper",
                                                            12 * 30))
                    self.accounts = self._load_accounts (node)
        elif toplevel != None:
            self.limit_nper = 12 * 30
            self.accounts = toplevel
            self.file = None
        else:
            raise RuntimeError

        self.filename = filename

    def get_accounts (self):
        return [self.accounts]

    def get_filename (self):
        return self.filename

    def set_filename (self, filename):
        self.filename = filename
        self.file = None

    def get_limit_nper (self):
        return self.limit_nper

    def set_limit_nper (self, limit_nper):
        self.limit_nper = limit_nper

    def is_readonly (self):
        return not isinstance (self.file, LockedFile)

    def get_xml (self):
        indent = " " * 2
        xml = "<money>\n%s<accounts limit-nper=\"%s\">\n" % (indent,
                                                             self.limit_nper)
        def save_account (account, depth):
            attrs = ""
            delay = account.get_delay ()
            if delay != 0:
                attrs += " delay=\"%s\"" % delay

            plot = account.get_plot ()
            if not plot:
                attrs += " plot=\"false\""

            sxml = ("\n%s<account name=\"%s\"%s>\n" %
                    (indent * depth,
                     account.get_name (),
                     attrs))
            acxml = account.write ()
            for line in acxml.split ('\n'):
                if len (line) > 0:
                    sxml += "%s%s\n" % (indent * (depth + 1), line)
            for each in account.accounts:
                sxml += save_account (each, depth + 1)
            sxml += "%s</account>\n" % (indent * depth)
            return sxml
            
        xml += save_account (self.accounts, 2)
        xml += "%s</accounts>\n</money>\n" % indent
        return xml

    def save (self, new_filename=None):
        xml = self.get_xml ()
        nodes = XML (xml)

        if new_filename != None:
            self.filename = new_filename
            self.file = None

        if self.file == None:
            self.file = LockedFile (self.filename, "w")

        content = tostring (nodes) + "\n"
        self.file.seek (0)
        self.file.write (content)
        self.file.flush ()
        self.file.truncate ()

    def _load_accounts (self, xml):
        accounts = None
        for node in xml.getchildren ():
            if node.tag == 'account':
                if accounts != None:
                    raise RuntimeError, "Too many top-level accounts"

                try:
                    accounts = self._load_account (node)
                except Exception, e:
                    print e

        return accounts

    def _load_account (self, xml):
        account = None
        accounts = []
        name = xml.attrib.get ('name', '')
        plot = xml.attrib.get ('plot', 'true')
        for node in xml.getchildren ():
            if node.tag == 'account':
                ac = self._load_account (node)
                if node.attrib.has_key ('delay'):
                    ac.set_delay (int (node.attrib['delay']))
                if node.attrib.has_key ('plot'):
                    f = node.attrib['plot']
                    if f.lower () == 'false':
                        ac.set_plot (False)
                accounts.append (ac)
            else:
                try:
                    data = create_account (node.tag, name)
                    if plot == 'false':
                        data.set_plot (False)
                    data.read (node)
                    account = data
                except Exception, e:
                    print "%s: %s" % (node.tag, e)
        account.accounts = accounts
        return account
